from collections import namedtuple
from datetime import datetime, timezone

from th2_data_services.interfaces.utils.converter import ITimestampConverter

_DatetimeTuple = namedtuple("DatetimeTuple", "datetime mantissa")


class DatetimeStringConverter(ITimestampConverter[str]):
    """Converts datetime strings.

    If you request microseconds but your timestamp has nanoseconds, they will be just cut (not rounding).

    Expected timestamp format "yyyy-MM-ddTHH:mm:ss[.SSSSSSSSS]Z".
    If you don't provide 'Z' in the end, it can return wrong results.
    """

    @classmethod
    def parse_timestamp(cls, datetime_string: str) -> (str, str):
        # Exception handling works faster than using `if`.
        try:
            # Handles "yyyy-MM-ddTHH:mm:ss.SSSSSSSSSZ"
            dt_tuple = _DatetimeTuple(*datetime_string.rsplit("."))
            timestamp = datetime.strptime(dt_tuple.datetime, "%Y-%m-%dT%H:%M:%S").replace(tzinfo=timezone.utc)
        except TypeError:
            # Handles "yyyy-MM-ddTHH:mm:ssZ"
            timestamp = datetime.strptime(datetime_string, "%Y-%m-%dT%H:%M:%SZ").replace(tzinfo=timezone.utc)
            dt_tuple = _DatetimeTuple("", "")  # ('2022-03-05T23:56:44', '0Z')

        mantissa_wo_z = dt_tuple.mantissa[:-1]
        nanoseconds = f"{mantissa_wo_z:0<9}"
        seconds = str(int(timestamp.timestamp()))

        return seconds, nanoseconds


class DatetimeConverter(ITimestampConverter[datetime]):
    """Converts datetime objects to timestamp.

    If you request milliseconds but your timestamp has microseconds, they will be just cut (not rounding).
    If you request nanoseconds, last 3 number will be zeros, because datatime object doesn't have nanoseconds.

    Expected timestamp format "datetime.datetime object".
    Expected that you provide UTC time in your data object.
    """

    @classmethod
    def parse_timestamp(cls, datetime_obj: datetime) -> (str, str):
        sec_and_mantissa = str(datetime_obj.replace(tzinfo=timezone.utc).timestamp()).split(".")
        seconds = sec_and_mantissa[0]
        nanoseconds = f"{sec_and_mantissa[1]:0<9}"
        return seconds, nanoseconds


class ProtobufTimestampConverter(ITimestampConverter[dict]):
    """Converts Th2 timestamps.

    If you request microseconds but your timestamp has nanoseconds, they will be just cut (not rounding).

    Expected timestamp format {'epochSecond': 123, 'nano': 500}.
    """

    @classmethod
    def parse_timestamp(cls, timestamp: dict) -> (str, str):
        seconds, nanoseconds = timestamp["epochSecond"], f"{timestamp['nano']:0>9}"
        return seconds, nanoseconds



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from collections import defaultdict
from typing import List, Dict, Optional, Union, Generator, Tuple, Callable

from treelib import Node, Tree
from treelib.exceptions import NodeIDAbsentError

from th2_data_services import Data
from th2_data_services.events_tree.events_tree import EventsTree
from th2_data_services.events_tree.events_tree import Th2Event
from th2_data_services.events_tree.exceptions import EventIdNotInTree
from th2_data_services.interfaces import IDataSource

# LOG import logging

import warnings

# LOG logger = logging.getLogger(__name__)


# LOG class _EventsTreeCollectionLogger(logging.LoggerAdapter):
# LOG     def process(self, msg, kwargs):
# LOG         return "ETC[%s] %s" % (self.extra["id"], msg), kwargs


class EventsTreeCollection(ABC):
    """EventsTreeCollection objective is building 'EventsTree's and storing them.

    - EventsTreeCollection stores all EventsTree. You can to refer to each of them.
    - Recovery of missing events occurs when you have passed DataSource class to constructor.
    Otherwise you should execute the method 'recover_unknown_events' manually.
    Note that there is no point in the method if the list of detached events is empty.
    """

    def __init__(
        self,
        data: Data,
        data_source: IDataSource = None,
        preserve_body: bool = False,
        stub: bool = False,
    ):
        """EventsTreeCollection constructor.

        Args:
            data: Data object with events.
            data_source: Provider Data Source object.
            preserve_body: If True it will preserve 'body' field in the Events.
            stub: If True it will create stub when event is broken.
        """
        self._id = id(self)
        self._preserve_body = preserve_body
        self._roots: List[EventsTree] = []
        self._parentless: Optional[List[EventsTree]] = None
        self._detached_nodes: Dict[Optional[str], List[Node]] = defaultdict(list)  # {parent_event_id: [Node1, ..]}
        self._stub_status = stub
        self._data_source = data_source
        # LOG         self._logger = _EventsTreeCollectionLogger(logger, {"id": self._id})

        events_nodes = self._build_event_nodes(data)
        self._build_trees(events_nodes)

        if data_source is not None:
            self.recover_unknown_events(self._data_source)

        if self._detached_nodes:
            w = "The collection were built with detached events because there are no some events in the source"
            # LOG             self._logger.warning(w)
            warnings.warn(w)

    def get_parentless_trees(self) -> List[EventsTree]:
        """Builds and returns parentless trees by detached events.

        Detached events will be removed from the tree.

        Returns:
            List of parentless trees if they exist, otherwise empty list.
        """
        if self._parentless is not None:
            return self._parentless
        else:
            self._parentless = self._build_parentless_trees()
            return self._parentless

    def _build_parentless_trees(self) -> List[EventsTree]:
        """Builds parentless trees by detached events.

        Returns:
            Parentless trees.
        """
        self._parentless = []

        stub_roots = list(self._detached_nodes.keys())
        for nodes in self._detached_nodes.values():
            for node in nodes:
                if node.identifier in stub_roots:
                    stub_roots.remove(node.identifier)

        for id_ in stub_roots:
            tree = Tree()
            tree.create_node(tag="Stub", identifier=id_, data=self._build_stub_event(id_))
            self._fill_tree(self._detached_nodes, tree, id_)
            self._parentless.append(EventsTree(tree))

        return self._parentless

    @abstractmethod
    def _build_stub_event(self, id_: str) -> dict:
        """Builds stub event.

        Args:
            id_: Event Id.
        """

    def _build_event_nodes(self, data: Data) -> Dict[Optional[str], List[Node]]:
        """Builds event nodes and group them by parent_event_id.

        Args:
            data: Data.

        Returns:
            Nodes.
        """
        events_nodes: Dict[Optional[str], List[Node]] = defaultdict(list)  # {parent_event_id: [Node1, Node2, ..]}

        for event in self._parse_events(data):
            parent_event_id: str = self._get_parent_event_id(event)
            node = self._build_node(event)
            events_nodes[parent_event_id].append(node)

        return events_nodes

    def _build_trees(self, nodes: Dict[Optional[str], List[Node]]) -> None:
        """Builds trees and saves detached events.

        Args:
            nodes: Events nodes.
        """
        roots = []
        for root_node in nodes[None]:  # None - is parent_event_id for root events.
            tree = Tree()
            tree.add_node(root_node)
            roots.append(tree)
            root_event_id: str = root_node.identifier
            self._fill_tree(nodes, tree, root_event_id)
        nodes.pop(None)

        self._roots = [EventsTree(root) for root in roots]

        self._detached_nodes = nodes

    def _fill_tree(self, nodes: Dict[Optional[str], List[Node]], current_tree: Tree, parent_id: str) -> None:
        """Fills tree recursively.

        Args:
            nodes: Events nodes.
            current_tree: Tree for fill.
            parent_id: Parent even id.
        """
        for node in nodes[parent_id].copy():
            event_id: str = node.identifier
            if event_id not in current_tree:
                current_tree.add_node(node, parent=parent_id)
            nodes[parent_id].remove(node)
            self._fill_tree(nodes, current_tree, event_id)  # Recursive fill.
        nodes.pop(parent_id)

    def _parse_event(self, event: Th2Event) -> Th2Event:
        """Parses event.

        Args:
            event: Event.

        Returns:
            Parsed event.
        """
        if not self._preserve_body:
            try:
                event.pop("body")
            except KeyError:
                pass
        return event

    def _parse_events(self, data: Union[Data, List[dict]]) -> Generator[dict, None, None]:
        """Parses events.

        Args:
            data: Data.

        Returns:
            Parsed events iterator.
        """
        for event in data:
            event = self._parse_event(event)
            yield event

    def _build_node(self, event: dict) -> Node:
        """Builds event as dict into event as Node.

        Args:
            event: Event.

        Returns:
            Node.
        """
        event_id: str = self._get_event_id(event)
        event_name: str = self._get_event_name(event)

        node = Node(tag=event_name, identifier=event_id, data=event)
        return node

    def append_event(self, event: dict) -> None:
        """Appends event into a tree.

        Args:
            event: Event.
        """
        event: dict = self._parse_event(event)
        node: Node = self._build_node(event)
        parent_event_id: str = self._get_parent_event_id(event)

        if parent_event_id is not None:
            events_trees = list(filter(lambda tree: parent_event_id in tree, self._roots))
            if events_trees:
                event_tree = events_trees[0]
                if node.identifier in event_tree:
                    pass
                else:
                    event_tree._append_node(node, parent_event_id)
                    self._fill_tree(self._detached_nodes, event_tree._tree, parent_event_id)
            else:
                self._detached_nodes[parent_event_id].append(node)
        else:
            tree = Tree()
            tree.add_node(node)
            self._roots.append(EventsTree(tree))
            self._fill_tree(self._detached_nodes, tree, node.identifier)

    @property
    def detached_events(self) -> dict:
        """Returns detached events as a dict that looks like {'parent_id': ['referenced event', ...]}."""
        w = (
            f"This method is deprecated since v1.1.0, and will be removed in the future. "
            f"Please use get_detached_events_iter or get_detached_events."
        )
        warnings.warn(message=w, category=DeprecationWarning, stacklevel=2)
        return self._detached_events()

    def _detached_events(self) -> dict:
        """Returns detached events as a dict that looks like {'parent_id': ['referenced event', ...]}."""
        return {id_: [node.data for node in nodes] for id_, nodes in self._detached_nodes.items()}

    def get_detached_events_iter(self) -> Generator[Th2Event, None, None]:
        """Yields detached events."""
        for nodes in self._detached_nodes.values():
            for node in nodes:
                yield node.data

    def get_detached_events(self) -> List[Th2Event]:
        """Returns detached events list."""
        return list(self.get_detached_events_iter())

    def get_roots_ids(self) -> List[str]:
        """Returns ids of all trees roots located in the collection.

        If there are parentless trees, they also will be return.
        """
        if self._parentless is not None:
            return [tree.get_root_id() for tree in self._roots] + [tree.get_root_id() for tree in self._parentless]
        return [tree.get_root_id() for tree in self._roots]

    def get_trees(self) -> List[EventsTree]:
        """Returns the list of trees inside the collection.

        If there are parentless trees, they also will be return.
        """
        if self._parentless is not None:
            return self._roots + self._parentless
        return self._roots

    def get_root_by_id(self, id) -> Th2Event:
        """Returns the root event of a tree by id of any event in this tree.

        If event id of parentless tree is passed, stub of this parentless tree will be returnd.

        Args:
            id: Event id.

        Returns:
            Th2Event.

        Raises:
            EventIdNotInTree: If event id is not in the trees.
        """
        try:
            return self.get_tree_by_id(id).get_root()
        except EventIdNotInTree:
            raise EventIdNotInTree(id)

    def get_tree_by_id(self, id) -> EventsTree:
        """Returns a tree by id of any event in this tree.

        If event id of parentless tree is passed, stub of this parentless tree will be returnd.

        Args:
            id: Event id.

        Returns:
            Tree.

        Raises:
            EventIdNotInTree: If event id is not in the trees.
        """
        for tree in self._roots:
            if id in tree:
                return tree
        if self._parentless is not None:
            for tree in self._parentless:
                if id in tree:
                    return tree
        raise EventIdNotInTree(id)

    def show(self):
        """Prints all EventsTrees as tree view.

        For example:

        ```
        Root
            |___ C01
            |    |___ C11
            |         |___ C111
            |         |___ C112
            |___ C02
            |___ C03
            |    |___ C31
        ```
        """
        trees = self.get_trees()
        for tree in trees:
            tree.show()

    @abstractmethod
    def _get_parent_event_id(self, event):
        """Gets parent event id from the event."""

    @abstractmethod
    def _get_event_id(self, event):
        """Gets event id from the event."""

    @abstractmethod
    def _get_event_name(self, event):
        """Gets event name from the event."""

    @abstractmethod
    def _get_events_by_id_resolver(self) -> Callable:
        """Gets a function that solve which protocol command to choose."""

    def __len__(self) -> int:
        """Returns the number of all events, including detached events."""
        return self.len_trees + self.len_detached_events

    def __contains__(self, event_id: str):
        for tree in self._roots:
            if event_id in tree:
                return True
        if self._parentless is not None:
            for tree in self._parentless:
                if event_id in tree:
                    return True
        return False

    def __repr__(self) -> str:
        len_trees_events = self.len_trees
        len_detached_events = self.len_detached_events
        len_regular_trees = len(self.get_roots_ids())

        if self._parentless:
            len_parentless_trees = len(self.get_parentless_trees())
            trees_parentless_info = (
                f"[regular={len_regular_trees - len_parentless_trees}, parentless={len_parentless_trees}]"
            )
        else:
            trees_parentless_info = ""

        return (
            f"{self.__class__.__name__}(trees={len_regular_trees}{trees_parentless_info}, "
            f"events={len_trees_events + len_detached_events}[trees={len_trees_events}, detached={len_detached_events}])"
        )

    def summary(self) -> str:
        """Returns the collection summary.

        The same as repr(EventsTreeCollection).
        """
        return self.__repr__()

    @property
    def len_trees(self) -> int:
        """Returns number of events in the trees inside the collection, including parentless trees."""
        if self._parentless is not None:
            return sum([len(root) for root in self._roots]) + self.len_parentless
        return sum([len(root) for root in self._roots])

    @property
    def len_parentless(self) -> int:
        """Returns number of events in the parentless trees inside the collection."""
        if self._parentless is not None:
            return sum([len(root) for root in self._parentless])
        return 0

    @property
    def len_detached_events(self) -> int:
        """Returns number of detached events in the collection."""
        return sum([len(nodes_lst) for nodes_lst in self._detached_nodes.values()])

    def get_all_events_iter(self) -> Generator[Th2Event, None, None]:
        """Yields all events from the collection."""
        for tree in self._roots:
            yield from tree.get_all_events_iter()
        if self._detached_nodes:
            yield from self.get_detached_events_iter()
        if self._parentless is not None:
            for tree in self._parentless:
                yield from tree.get_all_events_iter()

    def get_all_events(self) -> List[Th2Event]:
        """Returns all events from the collection."""
        return list(self.get_all_events_iter())

    def get_event(self, id: str) -> Optional[Th2Event]:
        """Returns an event by its id.

        Args:
            id: Event id.

        Raises:
            EventIdNotInTree: If event id is not in the collection.
        """
        for tree in self._roots:
            try:
                return tree.get_event(id)
            except EventIdNotInTree:
                continue
        if self._detached_nodes:
            for event in self.get_detached_events_iter():
                if self._get_event_id(event) == id:
                    return event
        if self._parentless is not None:
            for tree in self._parentless:
                try:
                    return tree.get_event(id)
                except EventIdNotInTree:
                    continue
        raise EventIdNotInTree(id)

    def get_leaves(self) -> Tuple[Th2Event]:
        """Returns all trees leaves, including parentless trees."""
        return tuple(self.get_leaves_iter())

    def get_leaves_iter(self) -> Generator[Th2Event, None, None]:
        """Yields all trees leaves, including parentless trees."""
        for tree in self._roots:
            yield from tree.get_leaves_iter()
        if self._parentless is not None:
            for tree in self._parentless:
                yield from tree.get_leaves_iter()

    def get_children(self, id: str) -> Tuple[Th2Event]:
        """Returns children of the event by its id.

        This method applicable only for trees (regular or parentless), not for detached events.

        Args:
            id: Event id.

        Raises:
            EventIdNotInTree: If event id is not in the trees.
        """
        for tree in self._roots:
            try:
                return tree.get_children(id)
            except EventIdNotInTree:
                continue
        if self._parentless is not None:
            for tree in self._parentless:
                try:
                    return tree.get_children(id)
                except EventIdNotInTree:
                    continue
        raise EventIdNotInTree(id)

    def get_children_iter(self, id: str) -> Generator[Th2Event, None, None]:
        """Yields children of the event by its id.

        This method applicable only for trees (regular or parentless), not for detached events.

        Args:
            id: Event id.

        Raises:
            EventIdNotInTree: If event id is not in the trees.
        """
        is_iter = False
        for tree in self._roots:
            try:
                yield from tree.get_children_iter(id)
                is_iter = True
            except EventIdNotInTree:
                continue
        if self._parentless is not None:
            for tree in self._parentless:
                try:
                    yield from tree.get_children_iter(id)
                    is_iter = True
                except EventIdNotInTree:
                    continue
        if not is_iter:
            raise EventIdNotInTree(id)

    def get_parent(self, id: str) -> Th2Event:
        """Returns a parent of the event by its id.

        Args:
            id: Event id.

        Raises:
            NodeIDAbsentError: If event id is not in the trees.
        """
        for tree in self._roots:
            try:
                return tree.get_parent(id)
            except EventIdNotInTree:
                continue
        if self._detached_nodes:
            parent_id = None
            for event in self.get_detached_events_iter():
                if self._get_event_id(event) == id:
                    parent_id = self._get_parent_event_id(event)
                    break
            for event in self.get_detached_events_iter():
                if self._get_event_id(event) == parent_id:
                    return event
        if self._parentless is not None:
            for tree in self._parentless:
                try:
                    return tree.get_parent(id)
                except EventIdNotInTree:
                    continue
        raise EventIdNotInTree(id)

    def get_full_path(self, id: str, field: str = None) -> List[Union[str, Th2Event]]:  # noqa: D412
        """Returns the full path for the event by its id in the right order.

        This method applicable only for trees (regular or parentless), not for detached events.

        Examples:

        Imagine we have the following tree.

        ```
        Harry
        ├── Bill
        └── Jane
            ├── Diane
            │   └── Mary
            └── Mark
        ```

        ```
        collection.get_full_path('Jane', id)
        ['Harry-event-id', 'Jane-event-id']

        collection.get_full_path('Jane', name)
        ['Harry-event-name', 'Jane-event-name']

        collection.get_full_path('Jane')
        ['Harry-event', 'Jane-event']
        ```

        Args:
            id: Event id.
            field: Field of event.

        Returns:
            Full path of event.

        Raises:
            EventIdNotInTree: If event id is not in the trees.
        """
        for tree in self._roots:
            try:
                return tree.get_full_path(id, field)
            except EventIdNotInTree:
                continue
        if self._parentless is not None:
            for tree in self._parentless:
                try:
                    return tree.get_full_path(id, field)
                except EventIdNotInTree:
                    continue
        raise EventIdNotInTree(id)

    def get_ancestors(self, id: str) -> List[Th2Event]:
        """Returns all event's ancestors in right order.

        This method applicable only for trees (regular or parentless), not for detached events.

        Args:
            id: Event id.

        Returns:
            All event's ancestors.

        Raises:
            EventIdNotInTree: If event id is not in the trees.
        """
        for tree in self._roots:
            try:
                return tree.get_ancestors(id)
            except EventIdNotInTree:
                continue
        if self._parentless is not None:
            for tree in self._parentless:
                try:
                    return tree.get_ancestors(id)
                except EventIdNotInTree:
                    continue
        raise EventIdNotInTree(id)

    def find_ancestor(self, id: str, filter: Callable) -> Optional[Th2Event]:
        """Finds the ancestor of an event.

        This method applicable only for trees (regular or parentless), not for detached events.

        Args:
            id: Event id.
            filter: Filter function

        Returns:
            Ancestor of Event.
        """
        for tree in self._roots:
            ancestor = tree.find_ancestor(id, filter)
            if ancestor is not None:
                return ancestor
        if self._parentless is not None:
            for tree in self._parentless:
                ancestor = tree.find_ancestor(id, filter)
                if ancestor is not None:
                    return ancestor
        return None

    def findall_iter(
        self,
        filter: Callable,
        stop: Callable = None,
        max_count: int = None,
    ) -> Generator[Th2Event, None, None]:
        """Searches events matches as iterator.

        - The search uses 'filter' which is a filtering function.
        - Optionally, the search uses 'stop' which is a stopping function.
        If 'stop' function returns 'True' then search is complete.
        - 'max_count' is a parameter that limits the search to a specified count.

        This method applicable only for trees (regular or parentless), not for detached events.

        Args:
            filter: Filter function.
            stop: Stop function. If None searches for all nodes in the trees.
            max_count: Max count of matched events. Stops searching when `max_count` will be reached.

        Yields:
            Matching events.
        """
        for tree in self._roots:
            yield from tree.findall_iter(filter=filter, stop=stop, max_count=max_count)
        if self._parentless is not None:
            for tree in self._parentless:
                yield from tree.findall_iter(filter=filter, stop=stop, max_count=max_count)

    def findall(
        self,
        filter: Callable,
        stop: Callable = None,
        max_count: int = None,
    ) -> List[Th2Event]:
        """Searches events matches.

        - The search uses 'filter' which is a filtering function.
        - Optionally, the search uses 'stop' which is a stopping function.
        If 'stop' function returns 'True' then search is complete.
        - 'max_count' is a parameter that limits the search to a specified count.

        This method applicable only for trees (regular or parentless), not for detached events.

        Args:
            filter: Filter function.
            stop: Stop function. If None searches for all nodes in the trees.
            max_count: Max count of matched events. Stops searching when `max_count` will be reached.

        Returns:
            Matching events.
        """
        return list(self.findall_iter(filter=filter, stop=stop, max_count=max_count))

    def find(self, filter: Callable, stop: Callable = None) -> Optional[Th2Event]:
        """Searches the first event match.

        - The search uses 'filter' which is a filtering function.
        - Optionally, the search uses 'stop' which is a stopping function.
        If 'stop' function returns 'True' then search is complete.

        This method applicable only for trees (regular or parentless), not for detached events.

        Args:
            filter: Filter function.
            stop: Stop function. If None searches for all nodes in the trees.

        Returns:
            One matching event.
        """
        for tree in self._roots:
            event = tree.find(filter=filter, stop=stop)
            if event is not None:
                return event
        if self._parentless is not None:
            for tree in self._parentless:
                event = tree.find(filter=filter, stop=stop)
                if event is not None:
                    return event
        return None

    def get_subtree(self, id: str) -> "EventsTree":
        """Returns subtree of the event by its id.

        This method applicable only for trees (regular or parentless), not for detached events.

        Args:
            id: Event id.

        Returns:
            Subtree.

        Raises:
            EventIdNotInTree: If event id is not in the trees.
        """
        for tree in self._roots:
            try:
                return tree.get_subtree(id)
            except (EventIdNotInTree, NodeIDAbsentError):
                continue
        if self._parentless is not None:
            for tree in self._parentless:
                try:
                    return tree.get_subtree(id)
                except (EventIdNotInTree, NodeIDAbsentError):
                    continue
        raise EventIdNotInTree(id)

    def recover_unknown_events(self, data_source: IDataSource) -> None:
        """Loads missed events and recover events.

        Args:
            data_source: Data Source.
        """
        resolver = self._get_events_by_id_resolver()
        instance_command = resolver(data_source)

        previous_detached_events = list(self._detached_events().keys())
        while previous_detached_events:
            called_command = instance_command(self._detached_events().keys(), use_stub=self._stub_status)
            events = data_source.command(called_command)

            for event in events:
                if not self._get_event_name(event) == "Broken_Event":
                    self.append_event(event)

            if previous_detached_events == list(self._detached_events().keys()):
                break
            previous_detached_events = list(self._detached_events().keys())

    def get_parentless_tree_collection(self) -> "EventsTreeCollection":
        """Builds and returns parentless trees by detached events as EventsTreeCollection.

        Detached events will be removed from the collection.

        Returns:
            EventsTreeCollection.
        """
        new_etc = self.__class__(
            data=[], data_source=self._data_source, preserve_body=self._preserve_body, stub=self._stub_status
        )
        new_etc._roots = self.get_parentless_trees()
        return new_etc



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List, Tuple, Generator, Callable, Optional, Union

from treelib import Tree, Node
from treelib.exceptions import NodeIDAbsentError

from th2_data_services.events_tree.exceptions import EventIdNotInTree

Th2Event = dict


class EventsTree:
    """EventsTree is a tree-based data structure of events.

    - get_x methods raise Exceptions if no result is found.
    - find_x methods return None if no result is found.
    - EventsTree stores events as Nodes and interacts with them using an internal tree.
    - EventsTree removes the 'body' field by default to save memory, but you can keep it.
    - Note that EventsTree stores only one tree.
        If you want to store all trees, use EventsTreeCollections.
    - EventsTree contains all events in memory.

    Take a look at the following HTML tree to understand some important terms.

    ```
    <body> <!-- ancestor (grandparent), but not parent -->
        <div> <!-- parent & ancestor -->
            <p>Hello, world!</p> <!-- child -->
            <p>Goodbye!</p> <!-- sibling -->
        </div>
    </body>
    ```
    """

    def __init__(self, tree: Tree):
        """EventsTree constructor.

        Args:
            tree (treelib.Tree): Tree.
        """
        self._tree = tree

    def _append_node(self, node: Node, parent_id: str) -> None:
        """Appends a node to the tree.

        Args:
            node: Node.
            parent_id: Parent event id.
        """
        if parent_id not in self._tree:
            raise NodeIDAbsentError(f"Node {parent_id} is not in the tree.")

        self._tree.add_node(node, parent_id)

    def get_all_events_iter(self) -> Generator[Th2Event, None, None]:
        """Returns all events from the tree as iterator."""
        for node in self._tree.all_nodes_itr():
            yield node.data

    def get_all_events(self) -> List[Th2Event]:
        """Returns all events from the tree."""
        return [node.data for node in self._tree.all_nodes()]

    def get_event(self, id: str) -> Th2Event:
        """Returns an event by id.

        Args:
            id: Event id.

        Raises:
            EventIdNotInTree: If event id is not in the tree.
        """
        node: Node = self._tree.get_node(id)
        if node is None:
            raise EventIdNotInTree(id)
        return node.data

    # TODO: In future it will be added.
    # def __getitem__(self, item):
    #     pass

    def get_root_id(self) -> str:
        """Returns the root id."""
        return self._tree.root

    def get_root_name(self) -> str:
        """Returns the root name."""
        return self._tree.get_node(self.get_root_id()).tag

    def get_root(self) -> Th2Event:
        """Returns the root event."""
        return self.get_event(self._tree.root)

    def get_leaves(self) -> Tuple[Th2Event]:
        """Returns all tree leaves."""
        # Do not use self.get_leaves_iter here because it takes more time.
        return tuple(self.get_leaves_iter())

    def get_leaves_iter(self) -> Generator[Th2Event, None, None]:
        """Returns all tree leaves as iterator."""
        for leaf in self._tree.leaves():
            yield leaf.data

    def get_children(self, id: str) -> Tuple[Th2Event]:
        """Returns children for the event by its id.

        Args:
            id: Event id.

        Raises:
            EventIdNotInTree: If event id is not in the tree.
        """
        try:
            return tuple(child.data for child in self._tree.children(id))
        except NodeIDAbsentError:
            raise EventIdNotInTree(id)

    def get_children_iter(self, id: str) -> Generator[Th2Event, None, None]:
        """Returns children as iterator for the event by its id.

        Args:
            id: Event id.

        Raises:
            EventIdNotInTree: If event id is not in the tree.
        """
        try:
            for child in self._tree.children(id):
                yield child.data
        except NodeIDAbsentError:
            raise EventIdNotInTree(id)

    def get_parent(self, id: str) -> Th2Event:
        """Returns a parent for the event by its id.

        Args:
            id: Event id.

        Raises:
            EventIdNotInTree: If event id is not in the tree.
        """
        try:
            parent = self._tree.parent(id)

            if parent is None:
                raise EventIdNotInTree(id)

            return parent.data
        except NodeIDAbsentError:
            raise EventIdNotInTree(id)

    def get_full_path(self, id: str, field: str = None) -> List[Union[str, Th2Event]]:  # noqa: D412
        """Returns the full path for the event by its id in right order.

        Examples:

        Imagine we have the following tree.

        ```
        Harry
        ├── Bill
        └── Jane
            ├── Diane
            │   └── Mary
            └── Mark
        ```

        ```
        tree.get_full_path('Jane', id)
        ['Harry-event-id', 'Jane-event-id']

        tree.get_full_path('Jane', name)
        ['Harry-event-name', 'Jane-event-name']

        tree.get_full_path('Jane')
        ['Harry-event', 'Jane-event']
        ```

        Args:
            id: Event id.
            field: Field of event.

        Returns:
            Full path of event.

        Raises:
            EventIdNotInTree: If event id is not in the tree.
        """
        result = []

        for event in self.get_ancestors(id) + [self.get_event(id)]:
            if field is None:
                result.append(event)
            else:
                result.append(event[field])

        return result

    def get_ancestors(self, id: str) -> List[Th2Event]:
        """Returns all event's ancestors in right order.

        Args:
            id: Event id.

        Returns:
            All event's ancestors.

        Raises:
            EventIdNotInTree: If event id is not in the tree.
        """
        result = [e for e in self._iter_ancestors(id)]
        result.reverse()
        return result

    def _iter_ancestors(self, id: str) -> Generator[Th2Event, None, None]:
        """Search ancestors by event id.

        Note, it yields ancestors in reverse order.

        Args:
            id: Event id.

        Yields:
            Ancestor of event.

        Raises:
            EventIdNotInTree: If event id is not in the tree.
        """
        try:
            ancestor: Node = self._tree.parent(id)
        except NodeIDAbsentError:
            raise EventIdNotInTree(id)

        while ancestor:
            yield ancestor.data
            ancestor = self._tree.parent(ancestor.identifier)

    def find_ancestor(self, id: str, filter: Callable) -> Optional[Th2Event]:
        """Finds the ancestor of an event.

        Args:
            id: Event id.
            filter: Filter function

        Returns:
            Ancestor of Event.
        """
        try:
            for ancestor in self._iter_ancestors(id):
                if filter(ancestor):
                    return ancestor
        except EventIdNotInTree:
            return None
        return None

    def findall_iter(
        self,
        filter: Callable,
        stop: Callable = None,
        max_count: int = None,
    ) -> Generator[Th2Event, None, None]:
        """Searches events matches as iterator.

        - The search uses 'filter' which is a filtering function.
        - Optionally, the search uses 'stop' which is a stopping function.
        If 'stop' function returns 'True' then search is complete.
        - 'max_count' is a parameter that limits the search to a specified count.

        Args:
            filter: Filter function.
            stop: Stop function. If None searches for all nodes in the tree.
            max_count: Max count of matched events. Stops searching when `max_count` will be reached.

        Yields:
            Matching events.
        """
        counter = 0
        for node in self._tree.all_nodes_itr():
            event = node.data.copy()
            if stop is not None and stop(event):
                break
            if filter(event):
                yield event
                counter += 1
                if max_count is not None and max_count <= counter:
                    break

    def findall(
        self,
        filter: Callable,
        stop: Callable = None,
        max_count: int = None,
    ) -> List[Th2Event]:
        """Searches events matches.

        - The search uses 'filter' which is a filtering function.
        - Optionally, the search uses 'stop' which is a stopping function.
        If 'stop' function returns 'True' then search is complete.
        - 'max_count' is a parameter that limits the search to a specified count.

        Args:
            filter: Filter function.
            stop: Stop function. If None searches for all nodes in the tree.
            max_count: Max count of matched events. Stops searching when `max_count` will be reached.

        Returns:
            Matching events.
        """
        result = []
        for event in self.findall_iter(filter=filter, stop=stop, max_count=max_count):
            result.append(event)
        return result

    def find(self, filter: Callable, stop: Callable = None) -> Optional[Th2Event]:
        """Searches the first event match.

        - The search uses 'filter' which is a filtering function.
        - Optionally, the search uses 'stop' which is a stopping function.
        If 'stop' function returns 'True' then search is complete.

        Args:
            filter: Filter function.
            stop: Stop function. If None searches for all nodes in the tree.

        Returns:
            One matching event.
        """
        for node in self._tree.all_nodes_itr():
            event = node.data
            if stop is not None and stop(event):
                break
            if filter(event):
                return event
        return None

    def get_subtree(self, id: str) -> "EventsTree":
        """Returns subtree of the event by its id.

        Args:
            id: Event id.

        Returns:
            Subtree.

        Raises:
            EventIdNotInTree: If event id is not in the tree.
        """
        subtree = self._tree.subtree(id)
        if not subtree:
            raise EventIdNotInTree(id)
        return EventsTree(subtree)

    def show(self) -> None:
        """Prints the EventsTree as tree view.

        For example:

        ```
        Root
            |___ C01
            |    |___ C11
            |         |___ C111
            |         |___ C112
            |___ C02
            |___ C03
            |    |___ C31
        ```
        """
        self._tree.show()

    def __len__(self) -> int:
        return len(self._tree)

    def __contains__(self, event_id: str):
        return event_id in self._tree

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name='{self.get_root_name()}', root_id='{self.get_root_id()}', events={len(self)})"

    def summary(self) -> str:
        """Returns the tree summary."""
        return self.__repr__()



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations
from abc import ABC, abstractmethod

from th2_data_services.interfaces.command import ICommand
from th2_data_services.interfaces.source_api import ISourceAPI

#LOG import logging

#LOG logger = logging.getLogger(__name__)


class IDataSource(ABC):
    @abstractmethod
    def command(self, cmd: ICommand):
        pass

    @property
    @abstractmethod
    def source_api(self) -> ISourceAPI:
        pass



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from typing import List


class IStub(ABC):
    class RequiredField:
        """This class used for mark required fields in your template."""

    REQUIRED_FIELD = RequiredField()

    def __init__(self):
        """Stab interface."""
        self._required_fields: list = self._define_required_fields()

    def _define_required_fields(self) -> List[str]:
        req_fields = []
        for k, v in self.template.items():
            if v is self.REQUIRED_FIELD:
                req_fields.append(k)

        return req_fields

    def _check_req_fields(self, fields):
        for rf in self._required_fields:
            if rf not in fields:
                raise TypeError(f"Required field '{rf}' is absent in changed fields list ({fields})")

    def _build_by_template(self, fields) -> dict:
        template = self.template.copy()
        for k, v in fields.items():
            if k in template:
                template[k] = v
        return template

    def build(self, fields: dict) -> dict:
        """Builds a stub.

        Args:
            fields:

        Returns:
            TypeError: If required fields is absent in changed fields list.
        """
        self._check_req_fields(fields)
        return self._build_by_template(fields)

    @property
    @abstractmethod
    def template(self) -> dict:
        pass


class IEventStub(IStub):
    """Just to mark Event Stub class."""

    def __repr__(self):
        return self.__class__.__name__


class IMessageStub(IStub):
    """Just to mark Message Stub class."""

    def __repr__(self):
        return self.__class__.__name__



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from typing import Any, Iterable


class IStreamAdapter(ABC):
    """Interface of Adapter for streams."""

    @abstractmethod
    def handle(self, stream: Iterable) -> Any:
        """Stream handle function that should yield data (not return)."""


class IRecordAdapter(ABC):
    """Interface of Adapter for record."""

    @abstractmethod
    def handle(self, record: dict) -> Any:
        """One record handle function that should return data."""



from abc import ABC, abstractmethod
from datetime import datetime, timezone
from typing import Generic, TypeVar

TimestampType = TypeVar("TimestampType")


class ITimestampConverter(ABC, Generic[TimestampType]):
    @classmethod
    @abstractmethod
    def parse_timestamp(cls, timestamp: TimestampType) -> (str, str):
        """Returns string representation of Unix time separated to seconds and nanoseconds.

        e.g. 2022-03-05T23:56:44.00123Z -> ('1646524604', '001230000')
        """

    @classmethod
    def to_datetime(cls, timestamp: TimestampType) -> datetime:
        """Converts timestamp to datetime object.

        If your timestamp has nanoseconds, they will be just cut (not rounding).

        Args:
            timestamp: TimestampType object to convert.

        Returns:
            datetime: Timestamp in python datetime format.
        """
        seconds, nanoseconds = cls.parse_timestamp(timestamp)
        microseconds = nanoseconds[:-3]
        return datetime.utcfromtimestamp(int(seconds)).replace(tzinfo=timezone.utc, microsecond=int(microseconds))

    @classmethod
    def to_microseconds(cls, timestamp: TimestampType) -> int:
        """Converts timestamp to microseconds.

        If your timestamp has nanoseconds, they will be just cut (not rounding).

        Args:
            timestamp: TimestampType object to convert.

        Returns:
            int: Timestamp in microseconds format.
        """
        seconds, nanoseconds = cls.parse_timestamp(timestamp)
        return int(f"{seconds}{nanoseconds[:-3]}")

    @classmethod
    def to_milliseconds(cls, timestamp: TimestampType):
        """Converts timestamp to milliseconds.

        If your timestamp has nanoseconds, they will be just cut (not rounding).

        Args:
            timestamp: TimestampType object to convert.

        Returns:
            int: Timestamp in microseconds format.
        """
        seconds, nanoseconds = cls.parse_timestamp(timestamp)
        return int(f"{seconds}{nanoseconds[:-6]}")

    @classmethod
    def to_nanoseconds(cls, timestamp: TimestampType) -> int:
        """Converts timestamp to nanoseconds.

        Args:
            timestamp: TimestampType object to convert.

        Returns:
            int: Timestamp in nanoseconds format.
        """
        seconds, nanoseconds = cls.parse_timestamp(timestamp)
        return int(f"{seconds}{nanoseconds}")



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from th2_data_services.data import Data

# LOG import logging
# LOG from logging import NullHandler

# Set default logging handler to avoid "No handler found" warnings.
# LOG logging.getLogger(__name__).addHandler(NullHandler())

# INTERACTIVE_MODE - is a global variable that tells the library not to delete
# the Data cache file if data iteration is interrupted.
INTERACTIVE_MODE = False  # Script mode by default.


# LOG def add_stderr_logger(level=logging.DEBUG):
# LOG     """Helper for quickly adding a StreamHandler to the logger.

# LOG     Useful for debugging.

# LOG     Returns the handler after adding it.
# LOG     """
# LOG     # This method needs to be in this __init__.py to get the __name__ correct
# LOG     # even if the lib is vendored within another package.
# LOG     logger = logging.getLogger(__name__)
# LOG     handler = logging.StreamHandler()
# LOG     handler.setFormatter(logging.Formatter("%(asctime)s %(levelname)s %(name)s : %(message)s"))
# LOG     logger.addHandler(handler)
# LOG     logger.setLevel(level)
# LOG     logger.debug("Added a stderr logging handler to logger: %s", __name__)
# LOG     return handler


# LOG def add_file_logger(filename="dslib.log", mode="w", level=logging.DEBUG):
# LOG     """Helper for quickly adding a StreamHandler to the logger.
# LOG
# LOG     Useful for debugging.
# LOG
# LOG     Returns the handler after adding it.
# LOG     """
# LOG     # This method needs to be in this __init__.py to get the __name__ correct
# LOG     # even if the lib is vendored within another package.
# LOG     logger = logging.getLogger(__name__)
# LOG     handler = logging.FileHandler(filename, mode=mode)
# LOG     handler.setFormatter(logging.Formatter("%(asctime)s %(levelname)s %(name)s : %(message)s"))
# LOG     logger.addHandler(handler)
# LOG     logger.setLevel(level)
# LOG     logger.debug("Added a file logging handler to logger: %s", __name__)
# LOG     return handler
